"""
Stencil quantity of interest (QoI) error bound safeguard.
"""

__all__ = ["StencilQuantityOfInterestErrorBoundSafeguard"]

from collections.abc import Sequence, Set
from functools import reduce
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _ones,
    _place,
    _reshape,
    _sliding_window_view,
    _zeros,
)
from ....utils.bindings import Bindings, Parameter
from ....utils.cast import (
    ToFloatMode,
    lossless_cast,
    saturating_finite_float_cast,
    to_float,
)
from ....utils.error import TypeCheckError, ctx, lookup_enum_or_raise
from ....utils.intervals import Interval, IntervalUnion
from ....utils.typing import JSON, F, S, T
from ..._qois import StencilQuantityOfInterest
from ..._qois.interval import compute_safe_data_lower_upper_interval_union
from ...eb import (
    ErrorBound,
    _apply_finite_qoi_error_bound,
    _check_error_bound,
    _compute_finite_absolute_error,
    _compute_finite_absolute_error_bound,
)
from ...qois import StencilQuantityOfInterestExpression
from .. import (
    BoundaryCondition,
    NeighbourhoodAxis,
    NeighbourhoodBoundaryAxis,
    _pad_with_boundary,
    _reverse_neighbourhood_indices,
)
from ..abc import StencilSafeguard


class StencilQuantityOfInterestErrorBoundSafeguard(StencilSafeguard):
    """
    The `StencilQuantityOfInterestErrorBoundSafeguard` guarantees that the
    pointwise error `type` on a derived quantity of interest (QoI) over a
    neighbourhood of data points is less than or equal to the provided bound
    `eb`.

    The quantity of interest is specified as a non-constant expression, in
    string form, over the neighbourhood tensor `X` that is centred on the
    pointwise value `x`. For example, to bound the error on the four-neighbour
    box mean in a 3x3 neighbourhood (where `x = X[I]`), set
    `qoi="(X[I[0]-1, I[1]] + X[I[0]+1, I[1]] + X[I[0], I[1]-1] + X[I[0], I[1]+1]) / 4"`.
    Note that `X` can be indexed absolute or relative to the centred data point
    `x` using the index array `I`.

    /// details | Tip: Finite Differences
        type: tip
    The stencil QoI safeguard can also be used to bound the pointwise error of
    the finite-difference-approximated derivative (of arbitrary order,
    accuracy, and grid spacing) over the data by using the `finite_difference`
    function in the `qoi` expression.
    ///

    /// details | Tip: Monotonic Sequences
        type: tip
    The stencil QoI safeguard can also be used to preserve the monotonicity of
    a sequence of values, i.e. to guarantee that a sequence that was originally
    strictly/weakly monotonically increasing/decreasing/constant still is. The
    sequence can be arbitrary within the stencil neighbourhood, e.g. along a
    single axis, in a zigzag, etc. Preserving the monotonicity of multiple
    sequences, e.g. along several axes, requires multiple stencil QoI
    safeguards. For instance, to guarantee that strictly increasing/decreasing
    sequences along a single axis stay strictly increasing/decreasing, use the
    following `qoi` expression with an absolute error bound of zero (more
    monotonicity QoIs, including strict vs weak monotonicity and constant
    sequences, can be found in [test_monotonicity.py]):

    [test_monotonicity.py]: https://github.com/juntyr/compression-safeguards/blob/main/tests/test_monotonicity.py

    <!--pytest.mark.skip-->
    ```py
    all([
        # strictly decreasing sequences stay strictly decreasing
        all(X[1:] < X[:-1]) == all(C["$X"][1:] < C["$X"][:-1]),
        # strictly increasing sequences stay strictly increasing
        all(X[1:] > X[:-1]) == all(C["$X"][1:] > C["$X"][:-1]),
    ])
    ```
    ///

    The shape of the data neighbourhood is specified as an ordered list of
    unique data axes and boundary conditions that are applied to these axes.
    If the safeguard is applied to data with an insufficient number of
    dimensions, it raises an exception. If the safeguard is applied to data
    with additional dimensions, it is indendently applied along these extra
    axes. For instance, a 2d QoI is applied to independently to all 2d slices
    in a 3d data cube.

    If the data neighbourhood uses the [valid][....BoundaryCondition.valid]
    boundary condition along an axis, the safeguard is only applied to data
    neighbourhoods centred on data points that have sufficient points before
    and after to satisfy the neighbourhood shape, i.e. it is not applied to
    all data points. If the axis is smaller than required by the neighbourhood
    along this axis, the safeguard is not applied. Using a different
    [`BoundaryCondition`][....BoundaryCondition] ensures that the safeguard is
    always applied to all data points.

    If the derived quantity of interest for a data neighbourhood evaluates to
    an infinite value, this safeguard guarantees that the quantity of interest
    on the corrected data neighbourhood produces the exact same infinite value.
    For a NaN quantity of interest, this safeguard guarantees that the quantity
    of interest on the corrected data neighbourhood is also NaN, but does not
    guarantee that it has the same bit pattern.

    The error bound can be verified by evaluating the QoI in the floating-point
    data type selected by `qoi_dtype` parameter using the
    [`evaluate_qoi`][.evaluate_qoi] method.

    Please refer to the
    [`StencilQuantityOfInterestExpression`][.....qois.StencilQuantityOfInterestExpression]
    for the EBNF grammar that specifies the language in which the quantities of
    interest are written.

    The implementation was originally inspired by:

    > Pu Jiao, Sheng Di, Hanqi Guo, Kai Zhao, Jiannan Tian, Dingwen Tao, Xin
    Liang, and Franck Cappello. (2022). Toward Quantity-of-Interest Preserving
    Lossy Compression for Scientific Data. *Proceedings of the VLDB Endowment*.
    16, 4 (December 2022), 697-710. Available from:
    [doi:10.14778/3574245.3574255](https://doi.org/10.14778/3574245.3574255).

    Parameters
    ----------
    qoi : StencilExpr
        The non-constant expression for computing the derived quantity of
        interest over a neighbourhood tensor `X`.
    neighbourhood : Sequence[dict[str, JSON] | NeighbourhoodBoundaryAxis]
        The non-empty axes of the data neighbourhood for which the quantity of
        interest is computed. The neighbourhood window is applied independently
        over any additional axes in the data.

        The per-axis boundary conditions are applied to the data in their order
        in the neighbourhood, i.e. earlier boundary extensions can influence
        later ones.
    type : str | ErrorBound
        The type of error bound on the quantity of interest that is enforced by
        this safeguard.
    eb : int | float | str | Parameter
        The value of or late-bound parameter name for the error bound on the
        quantity of interest that is enforced by this safeguard.

        The error bound is applied relative to the values of the quantity of
        interest evaluated on the original data.

        If `eb` is a late-bound parameter, its late-bound value must be
        broadcastable to the shape of the data to be safeguarded, not the shape
        of the QoI that has been evaluated (even though it is only applied to
        the QoI). The [`expand_qoi_to_data_shape`][.expand_qoi_to_data_shape]
        method can be used to expand an error-bound from the QoI shape to the
        data shape.
    qoi_dtype : str | ToFloatMode
        The floating-point data type in which the quantity of interest is
        evaluated. By default, the smallest floating-point data type that can
        losslessly represent all input data values is chosen.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    SyntaxError
        if the `qoi` is not a valid stencil quantity of interest expression.
    ValueError
        if the `neighbourhood` is empty.
    ValueError
        if any `neighbourhood.axis` is not unique.
    ValueError
        if `type` does not name a valid error bound, or the `qoi_dtype` does
        not name a valid floating-point data type.
    ValueError
        if `eb` is an invalid error bound value for the error bound `type`.
    ...
        if instantiating a neighbourhood boundary axis raises an exception.
    """

    __slots__: tuple[str, ...] = (
        "_qoi",
        "_neighbourhood",
        "_type",
        "_eb",
        "_qoi_dtype",
        "_qoi_expr",
    )
    _qoi: StencilQuantityOfInterestExpression
    _neighbourhood: tuple[NeighbourhoodBoundaryAxis, ...]
    _type: ErrorBound
    _eb: int | float | Parameter
    _qoi_dtype: ToFloatMode
    _qoi_expr: StencilQuantityOfInterest

    kind: ClassVar[str] = "qoi_eb_stencil"

    def __init__(
        self,
        qoi: StencilQuantityOfInterestExpression,
        neighbourhood: Sequence[dict[str, JSON] | NeighbourhoodBoundaryAxis],
        type: str | ErrorBound,
        eb: int | float | str | Parameter,
        qoi_dtype: str | ToFloatMode = ToFloatMode.lossless,
    ) -> None:
        with ctx.safeguard(self):
            with ctx.parameter("qoi"):
                TypeCheckError.check_instance_or_raise(qoi, str)

            with ctx.parameter("neighbourhood"):
                TypeCheckError.check_instance_or_raise(neighbourhood, Sequence)

                all_axes: list[int] = []
                neighbourhood_: list[NeighbourhoodBoundaryAxis] = []
                for i, axis in enumerate(neighbourhood):
                    with ctx.index(i):
                        TypeCheckError.check_instance_or_raise(
                            axis, dict | NeighbourhoodBoundaryAxis
                        )
                        axis_ = (
                            axis
                            if isinstance(axis, NeighbourhoodBoundaryAxis)
                            else NeighbourhoodBoundaryAxis.from_config(axis)
                        )
                        if axis_.axis in all_axes:
                            raise ValueError("axis must be unique") | ctx
                        all_axes.append(axis_.axis)
                        neighbourhood_.append(axis_)
                self._neighbourhood = tuple(neighbourhood_)

                if len(self._neighbourhood) <= 0:
                    raise ValueError("must not be empty") | ctx

            with ctx.parameter("type"):
                TypeCheckError.check_instance_or_raise(type, str | ErrorBound)
                self._type = (
                    type
                    if isinstance(type, ErrorBound)
                    else lookup_enum_or_raise(ErrorBound, type)
                )

            with ctx.parameter("eb"):
                TypeCheckError.check_instance_or_raise(
                    eb, int | float | str | Parameter
                )
                if isinstance(eb, Parameter):
                    self._eb = eb
                elif isinstance(eb, str):
                    self._eb = Parameter(eb)
                else:
                    _check_error_bound(self._type, eb)
                    self._eb = eb

            with ctx.parameter("qoi_dtype"):
                TypeCheckError.check_instance_or_raise(qoi_dtype, str | ToFloatMode)
                self._qoi_dtype = (
                    qoi_dtype
                    if isinstance(qoi_dtype, ToFloatMode)
                    else lookup_enum_or_raise(ToFloatMode, qoi_dtype)
                )

            shape = tuple(axis.before + 1 + axis.after for axis in self._neighbourhood)
            I = tuple(axis.before for axis in self._neighbourhood)  # noqa: E741

            with ctx.parameter("qoi"):
                self._qoi = qoi
                self._qoi_expr = StencilQuantityOfInterest(
                    qoi, stencil_shape=shape, stencil_I=I
                )

    @property
    @override
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        parameters = set(self._qoi_expr.late_bound_constants)

        if isinstance(self._eb, Parameter):
            parameters.add(self._eb)

        for axis in self._neighbourhood:
            if isinstance(axis.constant_boundary, Parameter):
                parameters.add(axis.constant_boundary)

        return frozenset(parameters)

    @override
    def compute_check_neighbourhood_for_data_shape(
        self,
        data_shape: tuple[int, ...],
    ) -> tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]:
        """
        Compute the shape of the data neighbourhood for data of a given shape.
        Boundary conditions of the same kind are combined, but separate kinds
        are tracked separately.

        An empty [`dict`][dict] is returned along dimensions for which the
        stencil QoI safeguard does not need to look at adjacent data points.

        This method also checks that the data shape is compatible with the
        stencil QoI safeguard.

        Parameters
        ----------
        data_shape : tuple[int, ...]
            The shape of the data.

        Returns
        -------
        neighbourhood_shape : tuple[dict[BoundaryCondition, NeighbourhoodAxis], ...]
            The shape of the data neighbourhood.

        Raises
        ------
        IndexError
            if any `neighbourhood` axis is out of bounds in `data_shape`.
        IndexError
            if any `neighbourhood` axis is duplicate.
        """

        neighbourhood: list[dict[BoundaryCondition, NeighbourhoodAxis]] = [
            dict() for _ in data_shape
        ]

        all_axes: list[int] = []
        with ctx.safeguard(self), ctx.parameter("neighbourhood"):
            for i, axis in enumerate(self._neighbourhood):
                with ctx.index(i), ctx.parameter("axis"):
                    if (axis.axis >= len(data_shape)) or (axis.axis < -len(data_shape)):
                        raise (
                            IndexError(
                                f"{axis.axis} is out of bounds for array of shape {data_shape}"
                            )
                            | ctx
                        )
                    naxis = axis.axis if axis.axis >= 0 else len(data_shape) + axis.axis
                    if naxis in all_axes:
                        raise (
                            IndexError(
                                f"duplicate axis index {axis.axis}, normalised to {naxis}, for array of shape {data_shape}"
                            )
                            | ctx
                        )
                    all_axes.append(naxis)

                    neighbourhood[naxis][axis.boundary] = axis.shape

        if any(a == 0 for a in data_shape):
            return tuple(dict() for _ in data_shape)

        return tuple(neighbourhood)

    def evaluate_qoi(
        self,
        data: np.ndarray[S, np.dtype[T]],
        late_bound: Bindings,
    ) -> np.ndarray[tuple[int, ...], np.dtype[F]]:
        """
        Evaluate the derived quantity of interest on the `data` in the
        floating-point data type selected by the `qoi_dtype` parameter.

        The quantity of interest may have a different shape if the
        [valid][.....BoundaryCondition.valid] boundary condition is used along
        any axis.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the quantity of interest is evaluated.
        late_bound : Bindings
            Bindings for late-bound constants in the quantity of interest.

        Returns
        -------
        qoi : np.ndarray[tuple[int, ...], np.dtype[F]]
            Evaluated quantity of interest, in floating-point.

        Raises
        ------
        IndexError
            if any `neighbourhood` axis is out of bounds in `data`.
        IndexError
            if any `neighbourhood` axis is duplicate.
        TypeError
            if the `data` could not be losslessly cast to `qoi_dtype`.
        LateBoundParameterResolutionError
            if any `neighbourhood` `axis.constant_boundary` is late-bound but
            its late-bound parameter is not in `late_bound`.
        ValueError
            if any `neighbourhood` `axis.constant_boundary` is late-bound but
            not a scalar.
        TypeError
            if any `neighbourhood` `axis.constant_boundary` is floating-point
            but the `data` is integer.
        ValueError
            if any `neighbourhood` `axis.constant_boundary` could not be
            losslessly converted to the `data`'s type.
        LateBoundParameterResolutionError
            if any of the `qoi`'s late-bound constants is not contained in the
            bindings.
        ValueError
            if any late-bound constant could not be broadcast to the `data`'s
            shape.
        TypeError
            if any late-bound constant is floating-point but the `data` is
            integer.
        ValueError
            if not all values for all late-bound constants could be losslessly
            converted to the `data`'s type.
        """

        # check that the data shape is compatible with the neighbourhood shape
        self.compute_check_neighbourhood_for_data_shape(data.shape)

        empty_shape = list(data.shape)

        # if the neighbourhood is empty, e.g. because we are in valid mode and
        #  the neighbourhood shape exceeds the data shape, return empty
        for axis in self._neighbourhood:
            if axis.boundary == BoundaryCondition.valid:
                empty_shape[axis.axis] = max(
                    0, empty_shape[axis.axis] - axis.before - axis.after
                )

        with ctx.safeguard(self):
            with ctx.parameter("qoi_dtype"):
                ftype: np.dtype[F] = self._qoi_dtype.floating_point_dtype_for(
                    data.dtype
                )  # type: ignore

            if any(s == 0 for s in empty_shape):
                return _zeros(tuple(empty_shape), dtype=ftype)

            constant_boundaries: list[None | np.ndarray[tuple[()], np.dtype[T]]] = []
            with ctx.parameter("neighbourhood"):
                for i, axis in enumerate(self._neighbourhood):
                    with ctx.index(i), ctx.parameter("constant_boundary"):
                        constant_boundaries.append(
                            None
                            if axis.constant_boundary is None
                            else late_bound.resolve_ndarray_with_lossless_cast(
                                axis.constant_boundary, (), data.dtype
                            )
                            if isinstance(axis.constant_boundary, Parameter)
                            else lossless_cast(axis.constant_boundary, data.dtype)
                        )

            data_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = data
            for axis, axis_constant_boundary in zip(
                self._neighbourhood, constant_boundaries
            ):
                data_boundary = _pad_with_boundary(
                    data_boundary,
                    axis.boundary,
                    axis.before,
                    axis.after,
                    axis_constant_boundary,
                    axis.axis,
                )

            window = tuple(axis.before + 1 + axis.after for axis in self._neighbourhood)

            data_windows_float_: np.ndarray[tuple[int, ...], np.dtype[F]] = to_float(
                _sliding_window_view(
                    data_boundary,
                    window,
                    axis=tuple(axis.axis for axis in self._neighbourhood),
                    writeable=False,
                ),
                ftype=ftype,
            )
            qoi_shape: tuple[int, ...] = data_windows_float_.shape[: -len(window)]
            data_windows_float: np.ndarray[
                tuple[int, *tuple[int, ...]], np.dtype[F]
            ] = _reshape(data_windows_float_, (-1, *window))

            late_bound_constants: dict[
                Parameter, np.ndarray[tuple[int, *tuple[int, ...]], np.dtype[F]]
            ] = dict()
            with ctx.parameter("qoi"):
                for c in self._qoi_expr.late_bound_constants:
                    with ctx.parameter(c):
                        late_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = (
                            late_bound.resolve_ndarray_with_lossless_cast(
                                c, data.shape, data.dtype
                            )
                        )
                    for axis, axis_constant_boundary in zip(
                        self._neighbourhood, constant_boundaries
                    ):
                        late_boundary = _pad_with_boundary(
                            late_boundary,
                            axis.boundary,
                            axis.before,
                            axis.after,
                            axis_constant_boundary,
                            axis.axis,
                        )
                    late_windows_float_: np.ndarray[tuple[int, ...], np.dtype[F]] = (
                        to_float(
                            _sliding_window_view(
                                late_boundary,
                                window,
                                axis=tuple(axis.axis for axis in self._neighbourhood),
                                writeable=False,
                            ),
                            ftype=ftype,
                        )
                    )
                    late_windows_float: np.ndarray[
                        tuple[int, *tuple[int, ...]], np.dtype[F]
                    ] = _reshape(late_windows_float_, (-1, *window))
                    late_bound_constants[c] = late_windows_float

        qoi_: np.ndarray[tuple[int], np.dtype[F]] = self._qoi_expr.eval(
            data_windows_float,
            late_bound_constants,
        )
        qoi: np.ndarray[tuple[int, ...], np.dtype[F]] = qoi_.reshape(qoi_shape)
        return qoi

    @override
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check which elements in the `prediction` array satisfy the error bound
        for the quantity of interest over a neighbourhood on the `data`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        IndexError
            if any `neighbourhood` axis is out of bounds in `data`.
        IndexError
            if any `neighbourhood` axis is duplicate.
        TypeError
            if the `data` could not be losslessly cast to `qoi_dtype`.
        LateBoundParameterResolutionError
            if any `neighbourhood` `axis.constant_boundary` is late-bound but
            its late-bound parameter is not in `late_bound`.
        ValueError
            if any `neighbourhood` `axis.constant_boundary` is late-bound but
            not a scalar.
        TypeError
            if any `neighbourhood` `axis.constant_boundary` is floating-point
            but the `data` is integer.
        ValueError
            if any `neighbourhood` `axis.constant_boundary` could not be
            losslessly converted to the `data`'s type.
        LateBoundParameterResolutionError
            if any of the `qoi`'s late-bound constants is not contained in the
            bindings.
        ValueError
            if any late-bound constant could not be broadcast to the `data`'s
            shape.
        TypeError
            if any late-bound constant is floating-point but the `data` is
            integer.
        ValueError
            if not all values for all late-bound constants could be losslessly
            converted to the `data`'s type.
        LateBoundParameterResolutionError
            if the error bound `eb` is late-bound but its late-bound parameter
            is not in `late_bound`.
        ValueError
            if the late-bound `eb` could not be broadcast to the `data`'s
            shape.
        ValueError
            if the late-bound `eb` is non-finite, i.e. infinite or NaN, or an
            invalid error bound value for the error bound `type`.
        """

        # check that the data shape is compatible with the neighbourhood shape
        self.compute_check_neighbourhood_for_data_shape(data.shape)

        window = tuple(axis.before + 1 + axis.after for axis in self._neighbourhood)

        # if the neighbourhood is empty, e.g. because we are in valid mode and
        #  the neighbourhood shape exceeds the data shape, allow all values
        for axis, w in zip(self._neighbourhood, window):
            if data.shape[axis.axis] < w and axis.boundary == BoundaryCondition.valid:
                return _ones(data.shape, dtype=np.dtype(np.bool))
        if data.size == 0:
            return _ones(data.shape, dtype=np.dtype(np.bool))

        with ctx.safeguard(self):
            constant_boundaries: list[None | np.ndarray[tuple[()], np.dtype[T]]] = []
            with ctx.parameter("neighbourhood"):
                for i, axis in enumerate(self._neighbourhood):
                    with ctx.index(i), ctx.parameter("constant_boundary"):
                        constant_boundaries.append(
                            None
                            if axis.constant_boundary is None
                            else late_bound.resolve_ndarray_with_lossless_cast(
                                axis.constant_boundary, (), data.dtype
                            )
                            if isinstance(axis.constant_boundary, Parameter)
                            else lossless_cast(axis.constant_boundary, data.dtype)
                        )

            data_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = data
            prediction_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = prediction
            for axis, axis_constant_boundary in zip(
                self._neighbourhood, constant_boundaries
            ):
                data_boundary = _pad_with_boundary(
                    data_boundary,
                    axis.boundary,
                    axis.before,
                    axis.after,
                    axis_constant_boundary,
                    axis.axis,
                )
                prediction_boundary = _pad_with_boundary(
                    prediction_boundary,
                    axis.boundary,
                    axis.before,
                    axis.after,
                    axis_constant_boundary,
                    axis.axis,
                )

            with ctx.parameter("qoi_dtype"):
                ftype: np.dtype[np.floating] = self._qoi_dtype.floating_point_dtype_for(
                    data.dtype
                )

            data_windows_float_: np.ndarray[tuple[int, ...], np.dtype[np.floating]] = (
                to_float(
                    _sliding_window_view(
                        data_boundary,
                        window,
                        axis=tuple(axis.axis for axis in self._neighbourhood),
                        writeable=False,
                    ),
                    ftype=ftype,
                )
            )
            qoi_shape: tuple[int, ...] = data_windows_float_.shape[: -len(window)]
            prediction_windows_float_: np.ndarray[
                tuple[int, ...], np.dtype[np.floating]
            ] = to_float(
                _sliding_window_view(
                    prediction_boundary,
                    window,
                    axis=tuple(axis.axis for axis in self._neighbourhood),
                    writeable=False,
                ),
                ftype=ftype,
            )

            data_windows_float: np.ndarray[
                tuple[int, *tuple[int, ...]], np.dtype[np.floating]
            ] = _reshape(data_windows_float_, (-1, *window))
            prediction_windows_float: np.ndarray[
                tuple[int, *tuple[int, ...]], np.dtype[np.floating]
            ] = _reshape(prediction_windows_float_, (-1, *window))

            late_bound_constants: dict[
                Parameter,
                np.ndarray[tuple[int, *tuple[int, ...]], np.dtype[np.floating]],
            ] = dict()
            with ctx.parameter("qoi"):
                for c in self._qoi_expr.late_bound_constants:
                    with ctx.parameter(c):
                        late_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = (
                            late_bound.resolve_ndarray_with_lossless_cast(
                                c, data.shape, data.dtype
                            )
                        )
                    for axis, axis_constant_boundary in zip(
                        self._neighbourhood, constant_boundaries
                    ):
                        late_boundary = _pad_with_boundary(
                            late_boundary,
                            axis.boundary,
                            axis.before,
                            axis.after,
                            axis_constant_boundary,
                            axis.axis,
                        )
                    late_windows_float_: np.ndarray[
                        tuple[int, ...], np.dtype[np.floating]
                    ] = to_float(
                        _sliding_window_view(
                            late_boundary,
                            window,
                            axis=tuple(axis.axis for axis in self._neighbourhood),
                            writeable=False,
                        ),
                        ftype=ftype,
                    )
                    late_windows_float: np.ndarray[
                        tuple[int, *tuple[int, ...]], np.dtype[np.floating]
                    ] = _reshape(late_windows_float_, (-1, *window))
                    late_bound_constants[c] = late_windows_float

            valid_slice = [slice(None)] * data.ndim
            for axis in self._neighbourhood:
                if axis.boundary == BoundaryCondition.valid:
                    start = None if axis.before == 0 else axis.before
                    end = None if axis.after == 0 else -axis.after
                    valid_slice[axis.axis] = slice(start, end)

            # optimization: only evaluate the QoI where necessary
            if where is not True:
                where_flat = where[tuple(valid_slice)].flatten()
                data_windows_float = np.compress(where_flat, data_windows_float, axis=0)
                prediction_windows_float = np.compress(
                    where_flat, prediction_windows_float, axis=0
                )
                late_bound_constants = {
                    c: np.compress(where_flat, cv, axis=0)
                    for (c, cv) in late_bound_constants.items()
                }

            qoi_data: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                self._qoi_expr.eval(
                    data_windows_float,
                    late_bound_constants,
                )
            )
            qoi_prediction: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                self._qoi_expr.eval(
                    prediction_windows_float,
                    late_bound_constants,
                )
            )

            with ctx.parameter("eb"):
                eb_: np.ndarray[tuple[()] | tuple[int, ...], np.dtype[np.floating]] = (
                    late_bound.resolve_ndarray_with_saturating_finite_float_cast(
                        self._eb,
                        data.shape,  # for simplicity, we resolve to the data shape
                        ftype,
                    )
                    if isinstance(self._eb, Parameter)
                    else saturating_finite_float_cast(self._eb, ftype)
                )
                if isinstance(self._eb, Parameter):
                    _eb: Parameter = self._eb
                    with ctx.late_bound_parameter(_eb):
                        # and then truncate it
                        eb_ = self.truncate_data_to_qoi_shape(eb_)
                        _check_error_bound(self._type, eb_)
                eb: np.ndarray[tuple[()] | tuple[int], np.dtype[np.floating]] = (
                    eb_ if eb_.shape == () else eb_.flatten()  # type: ignore
                )
                if where is not True and eb.shape != ():
                    eb = np.compress(where_flat, eb, axis=0)

        finite_ok: np.ndarray[tuple[int], np.dtype[np.bool]] = np.less_equal(
            _compute_finite_absolute_error(self._type, qoi_data, qoi_prediction),
            _compute_finite_absolute_error_bound(self._type, eb, qoi_data),
        )

        windows_ok_: np.ndarray[tuple[int], np.dtype[np.bool]] = _ensure_array(
            finite_ok
        )
        np.equal(qoi_data, qoi_prediction, out=windows_ok_, where=np.isinf(qoi_data))
        np.isnan(qoi_prediction, out=windows_ok_, where=np.isnan(qoi_data))

        # the check succeeds where `where` is False
        windows_ok: np.ndarray[tuple[int], np.dtype[np.bool]]
        if where is True:
            windows_ok = windows_ok_
        else:
            windows_ok = _ones(where_flat.shape, np.dtype(np.bool))
            _place(windows_ok, where_flat, windows_ok_)

        # the check succeeds for boundary points that were excluded by a valid
        #  boundary
        ok: np.ndarray[S, np.dtype[np.bool]] = _ones(
            data.shape, dtype=np.dtype(np.bool)
        )
        ok[tuple(valid_slice)] = windows_ok.reshape(qoi_shape)

        return ok

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the error bound is upheld with respect
        to the quantity of interest over a neighbourhood on the `data`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the error bound is upheld.

        Raises
        ------
        IndexError
            if any `neighbourhood` axis is out of bounds in `data`.
        IndexError
            if any `neighbourhood` axis is duplicate.
        TypeError
            if the `data` could not be losslessly cast to `qoi_dtype`.
        LateBoundParameterResolutionError
            if any `neighbourhood` `axis.constant_boundary` is late-bound but
            its late-bound parameter is not in `late_bound`.
        ValueError
            if any `neighbourhood` `axis.constant_boundary` is late-bound but
            not a scalar.
        TypeError
            if any `neighbourhood` `axis.constant_boundary` is floating-point
            but the `data` is integer.
        ValueError
            if any `neighbourhood` `axis.constant_boundary` could not be
            losslessly converted to the `data`'s type.
        LateBoundParameterResolutionError
            if any of the `qoi`'s late-bound constants is not contained in the
            bindings.
        ValueError
            if any late-bound constant could not be broadcast to the `data`'s
            shape.
        TypeError
            if any late-bound constant is floating-point but the `data` is
            integer.
        ValueError
            if not all values for all late-bound constants could be losslessly
            converted to the `data`'s type.
        LateBoundParameterResolutionError
            if the error bound `eb` is late-bound but its late-bound parameter
            is not in `late_bound`.
        ValueError
            if the late-bound `eb` could not be broadcast to the `data`'s
            shape.
        ValueError
            if the late-bound `eb` is non-finite, i.e. infinite or NaN, or an
            invalid error bound value for the error bound `type`.
        """

        # check that the data shape is compatible with the neighbourhood shape
        self.compute_check_neighbourhood_for_data_shape(data.shape)

        window = tuple(axis.before + 1 + axis.after for axis in self._neighbourhood)

        # if the neighbourhood is empty, e.g. because we are in valid mode and
        #  the neighbourhood shape exceeds the data shape, allow all values
        for axis, w in zip(self._neighbourhood, window):
            if data.shape[axis.axis] < w and axis.boundary == BoundaryCondition.valid:
                return Interval.full_like(data).into_union()
        if data.size == 0:
            return Interval.full_like(data).into_union()

        with ctx.safeguard(self):
            constant_boundaries: list[None | np.ndarray[tuple[()], np.dtype[T]]] = []
            with ctx.parameter("neighbourhood"):
                for i, axis in enumerate(self._neighbourhood):
                    with ctx.index(i), ctx.parameter("constant_boundary"):
                        constant_boundaries.append(
                            None
                            if axis.constant_boundary is None
                            else late_bound.resolve_ndarray_with_lossless_cast(
                                axis.constant_boundary, (), data.dtype
                            )
                            if isinstance(axis.constant_boundary, Parameter)
                            else lossless_cast(axis.constant_boundary, data.dtype)
                        )

            data_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = data
            for axis, axis_constant_boundary in zip(
                self._neighbourhood, constant_boundaries
            ):
                data_boundary = _pad_with_boundary(
                    data_boundary,
                    axis.boundary,
                    axis.before,
                    axis.after,
                    axis_constant_boundary,
                    axis.axis,
                )

            with ctx.parameter("qoi_dtype"):
                ftype: np.dtype[np.floating] = self._qoi_dtype.floating_point_dtype_for(
                    data.dtype
                )

            data_windows_float_: np.ndarray[tuple[int, ...], np.dtype[np.floating]] = (
                to_float(
                    _sliding_window_view(
                        data_boundary,
                        window,
                        axis=tuple(axis.axis for axis in self._neighbourhood),
                        writeable=False,
                    ),
                    ftype=ftype,
                )
            )
            qoi_stencil_shape: tuple[int, ...] = data_windows_float_.shape
            data_windows_float: np.ndarray[
                tuple[int, *tuple[int, ...]], np.dtype[np.floating]
            ] = _reshape(data_windows_float_, (-1, *window))

            late_bound_constants: dict[
                Parameter,
                np.ndarray[tuple[int, *tuple[int, ...]], np.dtype[np.floating]],
            ] = dict()
            with ctx.parameter("qoi"):
                for c in self._qoi_expr.late_bound_constants:
                    with ctx.parameter(c):
                        late_boundary: np.ndarray[tuple[int, ...], np.dtype[T]] = (
                            late_bound.resolve_ndarray_with_lossless_cast(
                                c, data.shape, data.dtype
                            )
                        )
                    for axis, axis_constant_boundary in zip(
                        self._neighbourhood, constant_boundaries
                    ):
                        late_boundary = _pad_with_boundary(
                            late_boundary,
                            axis.boundary,
                            axis.before,
                            axis.after,
                            axis_constant_boundary,
                            axis.axis,
                        )
                    late_windows_float_: np.ndarray[
                        tuple[int, ...], np.dtype[np.floating]
                    ] = to_float(
                        _sliding_window_view(
                            late_boundary,
                            window,
                            axis=tuple(axis.axis for axis in self._neighbourhood),
                            writeable=False,
                        ),
                        ftype=ftype,
                    )
                    late_windows_float: np.ndarray[
                        tuple[int, *tuple[int, ...]], np.dtype[np.floating]
                    ] = _reshape(late_windows_float_, (-1, *window))
                    late_bound_constants[c] = late_windows_float

            valid_slice = [slice(None)] * data.ndim
            for axis in self._neighbourhood:
                if axis.boundary == BoundaryCondition.valid:
                    start = None if axis.before == 0 else axis.before
                    end = None if axis.after == 0 else -axis.after
                    valid_slice[axis.axis] = slice(start, end)

            # optimization: only evaluate the QoI and compute data bounds where
            #  necessary
            if where is not True:
                where_flat = where[tuple(valid_slice)].flatten()
                data_windows_float = np.compress(where_flat, data_windows_float, axis=0)
                late_bound_constants = {
                    c: np.compress(where_flat, cv, axis=0)
                    for (c, cv) in late_bound_constants.items()
                }

            data_qoi: np.ndarray[tuple[int], np.dtype[np.floating]] = (
                self._qoi_expr.eval(
                    data_windows_float,
                    late_bound_constants,
                )
            )

            with ctx.parameter("eb"):
                eb_: np.ndarray[tuple[()] | tuple[int, ...], np.dtype[np.floating]] = (
                    late_bound.resolve_ndarray_with_saturating_finite_float_cast(
                        self._eb,
                        data.shape,  # for simplicity, we resolve to the data shape
                        ftype,
                    )
                    if isinstance(self._eb, Parameter)
                    else saturating_finite_float_cast(self._eb, ftype)
                )
                if isinstance(self._eb, Parameter):
                    _eb: Parameter = self._eb
                    with ctx.late_bound_parameter(_eb):
                        eb_ = self.truncate_data_to_qoi_shape(
                            eb_
                        )  # and then truncate it
                        _check_error_bound(self._type, eb_)
                eb: np.ndarray[tuple[()] | tuple[int], np.dtype[np.floating]] = (
                    eb_ if eb_.shape == () else eb_.flatten()  # type: ignore
                )
                if where is not True and eb.shape != ():
                    eb = np.compress(where_flat, eb, axis=0)

        qoi_lower_upper: tuple[
            np.ndarray[tuple[int], np.dtype[np.floating]],
            np.ndarray[tuple[int], np.dtype[np.floating]],
        ] = _apply_finite_qoi_error_bound(
            self._type,
            eb,
            data_qoi,
        )
        qoi_lower, qoi_upper = qoi_lower_upper

        # compute the bounds in data space
        data_windows_float_lower_, data_windows_float_upper_ = (
            self._qoi_expr.compute_data_bounds(
                qoi_lower,
                qoi_upper,
                data_windows_float,
                late_bound_constants,
            )
        )

        # the data bounds can be arbitrary where `where` is False since they
        #  are later overridden by preserve_only_where
        data_windows_float_lower: np.ndarray[tuple[int, ...], np.dtype[np.floating]]
        data_windows_float_upper: np.ndarray[tuple[int, ...], np.dtype[np.floating]]
        if where is True:
            data_windows_float_lower = data_windows_float_lower_.reshape(
                qoi_stencil_shape
            )
            data_windows_float_upper = data_windows_float_upper_.reshape(
                qoi_stencil_shape
            )
        else:
            where_indices = np.nonzero(where_flat)[0].reshape(
                (-1, *tuple(1 for _ in window))
            )
            data_windows_float_lower = np.full(
                qoi_stencil_shape, -np.inf, dtype=ftype
            ).reshape(-1, *window)
            data_windows_float_upper = np.full(
                qoi_stencil_shape, np.inf, dtype=ftype
            ).reshape(-1, *window)
            np.put_along_axis(
                data_windows_float_lower,
                where_indices,
                data_windows_float_lower_,
                axis=0,
            )
            np.put_along_axis(
                data_windows_float_upper,
                where_indices,
                data_windows_float_upper_,
                axis=0,
            )
            data_windows_float_lower = data_windows_float_lower.reshape(
                qoi_stencil_shape
            )
            data_windows_float_upper = data_windows_float_upper.reshape(
                qoi_stencil_shape
            )

        # only contribute window elements that are used in the QoI
        window_used = _zeros(window, dtype=np.dtype(np.bool))
        for idxs in self._qoi_expr.data_indices:
            window_used[idxs] = True

        # compute the reverse: for each data element, which windows is it in
        # i.e. for each data element, which QoI elements does it contribute to
        #      and thus which data bounds affect it
        reverse_indices_windows = _reverse_neighbourhood_indices(
            data.shape,
            self._neighbourhood,
            window_used,
            True if where is True else where_flat,
        )

        # flatten the QoI data bounds and append an infinite value,
        #  which is indexed if an element did not contribute to the maximum
        #  number of windows
        data_windows_float_lower_flat = np.full(
            data_windows_float_lower.size + 1, -np.inf, dtype=ftype
        )
        data_windows_float_upper_flat = np.full(
            data_windows_float_upper.size + 1, np.inf, dtype=ftype
        )
        data_windows_float_lower_flat[:-1] = data_windows_float_lower.flatten()
        data_windows_float_upper_flat[:-1] = data_windows_float_upper.flatten()

        # for each data element, reduce over the data bounds that affect it
        # since some data elements may have no data bounds that affect them,
        #  e.g. because of the valid boundary condition, they may have infinite
        #  bounds
        data_float_lower: np.ndarray[S, np.dtype[np.floating]] = np.amax(
            data_windows_float_lower_flat[reverse_indices_windows], axis=1
        ).reshape(data.shape)
        data_float_upper: np.ndarray[S, np.dtype[np.floating]] = np.amin(
            data_windows_float_upper_flat[reverse_indices_windows], axis=1
        ).reshape(data.shape)

        # only preserve intervals where some (neighbouring) data point
        #  contributed (back) safety requirements
        # this takes the `where` condition, after overlapping stencils, into
        #  account
        return compute_safe_data_lower_upper_interval_union(
            data, data_float_lower, data_float_upper
        ).preserve_only_where(
            (  # type: ignore
                reverse_indices_windows[:, 0] < data_windows_float_lower.size
            )
            # where `where` is not False, at least the isnan stays isnan rule
            #  should always be honoured, anything else would be unexpected
            | (True if where is True else where.flatten())
        )

    @override
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        The footprint usually extends beyond `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

            Conceptually, `where` is applied to `footprint` at the end.

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        # check that the foot shape is compatible with the neighbourhood shape
        self.compute_check_neighbourhood_for_data_shape(foot.shape)

        window = tuple(axis.before + 1 + axis.after for axis in self._neighbourhood)

        # if the neighbourhood is empty, e.g. because we are in valid mode and
        #  the neighbourhood shape exceeds the data shape, the footprint is
        #  empty
        for axis, w in zip(self._neighbourhood, window):
            if foot.shape[axis.axis] < w and axis.boundary == BoundaryCondition.valid:
                return _zeros(foot.shape, dtype=np.dtype(np.bool))
        if foot.size == 0:
            return _zeros(foot.shape, dtype=np.dtype(np.bool))

        foot_boundary: np.ndarray[tuple[int, ...], np.dtype[np.bool]] = foot
        for axis in self._neighbourhood:
            foot_boundary = _pad_with_boundary(
                foot_boundary,
                axis.boundary,
                axis.before,
                axis.after,
                None
                if axis.constant_boundary is None
                else _zeros((), np.dtype(np.bool)),
                axis.axis,
            )

        foot_windows: np.ndarray[tuple[int, ...], np.dtype[np.bool]] = (
            _sliding_window_view(
                foot_boundary,
                window,
                axis=tuple(axis.axis for axis in self._neighbourhood),
                writeable=False,
            )
        )

        valid_slice = [slice(None)] * foot.ndim
        for axis in self._neighbourhood:
            if axis.boundary == BoundaryCondition.valid:
                start = None if axis.before == 0 else axis.before
                end = None if axis.after == 0 else -axis.after
                valid_slice[axis.axis] = slice(start, end)

        # only contribute window elements that are used in the QoI
        window_used = _zeros(window, dtype=np.dtype(np.bool))
        for idxs in self._qoi_expr.data_indices:
            window_used[idxs] = True

        # expand into the footprint
        footprint = np.zeros_like(foot)
        footprint[tuple(valid_slice)] = np.logical_or.reduce(
            (foot_windows & window_used.reshape(((1,) * foot.ndim) + window)).reshape(
                foot_windows.shape[: foot.ndim] + (-1,)
            ),
            axis=foot.ndim,
        )
        footprint &= where

        return footprint

    @override
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        The inverse footprint usually extends beyond `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

            Conceptually, `where` is applied to `foot` at the start.

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        # check that the foot shape is compatible with the neighbourhood shape
        self.compute_check_neighbourhood_for_data_shape(foot.shape)

        window = tuple(axis.before + 1 + axis.after for axis in self._neighbourhood)

        # if the neighbourhood is empty, e.g. because we are in valid mode and
        #  the neighbourhood shape exceeds the data shape, the inverse
        #  footprint is empty
        for axis, w in zip(self._neighbourhood, window):
            if foot.shape[axis.axis] < w and axis.boundary == BoundaryCondition.valid:
                return _zeros(foot.shape, dtype=np.dtype(np.bool))
        if foot.size == 0:
            return _zeros(foot.shape, dtype=np.dtype(np.bool))

        valid_slice = [slice(None)] * foot.ndim
        for axis in self._neighbourhood:
            if axis.boundary == BoundaryCondition.valid:
                start = None if axis.before == 0 else axis.before
                end = None if axis.after == 0 else -axis.after
                valid_slice[axis.axis] = slice(start, end)

        # only contribute window elements that are used in the QoI
        window_used = _zeros(window, dtype=np.dtype(np.bool))
        for idxs in self._qoi_expr.data_indices:
            window_used[idxs] = True

        # valid foot, which includes the where condition
        foot_valid = (foot & where)[tuple(valid_slice)]

        # compute the reverse: for each data element, which windows is it in
        # i.e. for each data element, which QoI elements does it contribute to
        #      and thus which foot elements affect it
        reverse_indices_windows = _reverse_neighbourhood_indices(
            foot.shape,
            self._neighbourhood,
            window_used,
            # optimization: only compute the reverse indices where the foot is
            #  True, since the others don't contribute anyways
            where_flat=foot_valid.flatten(),
        )

        window_size = reduce(lambda x, y: x * y, window, 1)

        # optimization: since we only compute reverse indices where the foot is
        #  True, we can more efficiently check which elements are part of the
        #  inverse footprint by checking if they have any valid reverse
        #  contribution indices
        valid_reverse_index = reverse_indices_windows != (foot_valid.size * window_size)

        inverse_footprint: np.ndarray[S, np.dtype[np.bool]] = np.logical_or.reduce(
            valid_reverse_index, axis=1
        ).reshape(foot.shape)

        return inverse_footprint

    def truncate_data_to_qoi_shape(
        self, data: np.ndarray[tuple[int, ...], np.dtype[T]]
    ) -> np.ndarray[tuple[int, ...], np.dtype[T]]:
        """
        Truncate the `data` array to the shape of the `qoi` array that would be evaluated from it.

        For neighbourhoods without any [valid][.....BoundaryCondition.valid]
        boundaries, this method simply copies the `data` array. Along axes with
        a [valid][.....BoundaryCondition.valid] boundary, the `data` is
        truncated before and after by as many elements as the boundary cuts off.

        This method is the *lossy* inverse of
        [`expand_qoi_to_data_shape`][..expand_qoi_to_data_shape].

        Parameters
        ----------
        data : np.ndarray[tuple[int, ...], np.dtype[T]]
            Data array to be truncated.

        Returns
        -------
        data_truncated : np.ndarray[tuple[int, ...], np.dtype[T]]
            Truncated data array.

        Raises
        ------
        IndexError
            if any `neighbourhood` axis is out of bounds in `data`.
        IndexError
            if any `neighbourhood` axis is duplicate.
        """

        data = _ensure_array(data)

        qoi_index = [slice(None, None) for _ in data.shape]

        all_axes: list[int] = []
        for axis in self._neighbourhood:
            if (axis.axis >= data.ndim) or (axis.axis < -data.ndim):
                raise (
                    IndexError(
                        f"axis index {axis.axis} is out of bounds for array of shape {data.shape}"
                    )
                    | ctx
                )
            naxis = axis.axis if axis.axis >= 0 else data.ndim + axis.axis
            if naxis in all_axes:
                raise (
                    IndexError(
                        f"duplicate axis index {axis.axis}, normalised to {naxis}, for array of shape {data.shape}"
                    )
                    | ctx
                )
            all_axes.append(naxis)

            if axis.boundary != BoundaryCondition.valid:
                continue

            qoi_index[naxis] = slice(
                axis.before, None if axis.after == 0 else -axis.after
            )

        return data[tuple(qoi_index)]

    def expand_qoi_to_data_shape(
        self, qoi: np.ndarray[tuple[int, ...], np.dtype[T]]
    ) -> np.ndarray[tuple[int, ...], np.dtype[T]]:
        """
        Expand the `qoi` array to the shape of the data it was evaluated from.

        The `qoi` array is assumed to have been produced by the
        [`evaluate_qoi`][..evaluate_qoi] method.

        For neighbourhoods without any [valid][.....BoundaryCondition.valid]
        boundaries, this method simply copies the `qoi` array. Along axes with
        a [valid][.....BoundaryCondition.valid] boundary, the `qoi` is expanded
        before and after by as many elements as the boundary cut off. These new
        elements are filled with zeros.

        This method can be used to produce the array for the late-bound `eb`
        parameter by first deriving the error bound from an evaluated QoI array
        (shape) and then expanding it to the data shape.

        This method is the *lossy* inverse of
        [`truncate_data_to_qoi_shape`][..truncate_data_to_qoi_shape].

        Parameters
        ----------
        qoi : np.ndarray[tuple[int, ...], np.dtype[T]]
            QoI array to be expanded.

        Returns
        -------
        qoi_expanded : np.ndarray[tuple[int, ...], np.dtype[T]]
            Zero-expanded QoI array.

        Raises
        ------
        IndexError
            if any `neighbourhood` axis is out of bounds in `qoi`.
        IndexError
            if any `neighbourhood` axis is duplicate.
        """

        qoi = _ensure_array(qoi)

        data_shape = list(qoi.shape)
        qoi_index = [slice(None, None) for _ in qoi.shape]

        all_axes: list[int] = []
        for axis in self._neighbourhood:
            if (axis.axis >= qoi.ndim) or (axis.axis < -qoi.ndim):
                raise (
                    IndexError(
                        f"axis index {axis.axis} is out of bounds for array of shape {qoi.shape}"
                    )
                    | ctx
                )
            naxis = axis.axis if axis.axis >= 0 else qoi.ndim + axis.axis
            if naxis in all_axes:
                raise (
                    IndexError(
                        f"duplicate axis index {axis.axis}, normalised to {naxis}, for array of shape {qoi.shape}"
                    )
                    | ctx
                )
            all_axes.append(naxis)

            if axis.boundary != BoundaryCondition.valid:
                continue

            data_shape[naxis] += axis.before + axis.after
            qoi_index[naxis] = slice(
                axis.before, None if axis.after == 0 else -axis.after
            )

        out: np.ndarray[tuple[int, ...], np.dtype[T]] = _zeros(
            tuple(data_shape), dtype=qoi.dtype
        )
        out[tuple(qoi_index)] = qoi
        return out

    @override
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        return dict(
            kind=type(self).kind,
            qoi=self._qoi,
            neighbourhood=[axis.get_config() for axis in self._neighbourhood],
            type=self._type.name,
            eb=self._eb,
            qoi_dtype=self._qoi_dtype.name,
        )



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _ensure_array,
    _is_sign_negative_number,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds, guarantee_arg_within_expr_bounds
from ..typing import F, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarSinh(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarSinh":
        return ScalarSinh(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.sinh,
            ScalarSinh,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.sinh(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and sinh(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.sinh(argv)

        # apply the inverse function to get the bounds on arg
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(argv, np.asinh(expr_lower))
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(argv, np.asinh(expr_upper))
        )

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in sinh(asinh(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.sinh(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.sinh(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"sinh({self._a!r})"


class ScalarCosh(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarCosh":
        return ScalarCosh(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.cosh,
            ScalarCosh,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.cosh(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and cosh(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.cosh(argv)

        # apply the inverse function to get the bounds on arg
        al = np.acosh(_maximum_zero_sign_sensitive(expr_lower, Xs.dtype.type(1)))
        au = np.acosh(expr_upper)

        # flip and swap the expr bounds to get the bounds on arg
        # cosh(...) cannot be less than 1, but
        #  - a > 0 and 1 < el <= eu -> al = el, au = eu
        #  - a < 0 and 1 < el <= eu -> al = -eu, au = -el
        #  - el <= 1 -> al = -eu, au = eu
        # TODO: an interval union could represent that the two sometimes-
        #       disjoint intervals in the future
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(al, copy=True)
        np.negative(
            au,
            out=arg_lower,
            where=(np.less_equal(expr_lower, 1) | _is_sign_negative_number(argv)),
        )
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(au, copy=True)
        np.negative(
            al,
            out=arg_upper,
            where=(np.greater(expr_lower, 1) & _is_sign_negative_number(argv)),
        )
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in cosh(acosh(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.cosh(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.cosh(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"cosh({self._a!r})"


class ScalarTanh(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarTanh":
        return ScalarTanh(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.tanh,
            ScalarTanh,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.tanh(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and tanh(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.tanh(argv)

        # apply the inverse function to get the bounds on arg
        # ensure that the bounds on tanh(...) are in [-1, +1]
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(
                argv,
                np.atanh(_maximum_zero_sign_sensitive(Xs.dtype.type(-1), expr_lower)),
            )
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(
                argv,
                np.atanh(_minimum_zero_sign_sensitive(expr_upper, Xs.dtype.type(1))),
            )
        )

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in tanh(atanh(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.tanh(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.tanh(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"tanh({self._a!r})"


class ScalarAsinh(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAsinh":
        return ScalarAsinh(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.asinh,
            ScalarAsinh,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.asinh(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and asinh(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.asinh(argv)

        # apply the inverse function to get the bounds on arg
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _minimum_zero_sign_sensitive(argv, np.sinh(expr_lower))
        )
        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            _maximum_zero_sign_sensitive(argv, np.sinh(expr_upper))
        )

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in asinh(sinh(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.asinh(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.asinh(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"asinh({self._a!r})"


class ScalarAcosh(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAcosh":
        return ScalarAcosh(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.acosh,
            ScalarAcosh,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.acosh(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and acosh(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.acosh(argv)

        eps_one = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(0))

        # apply the inverse function to get the bounds on arg
        # acosh(...) is NaN for values smaller than 1 and can then take any
        #  value smaller than one
        # otherwise ensure that the bounds on acosh(...) are non-negative
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.cosh(_maximum_zero_sign_sensitive(Xs.dtype.type(0), expr_lower))
        )
        arg_lower[np.less(argv, 1)] = -np.inf
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.cosh(expr_upper)
        )
        arg_upper[np.less(argv, 1)] = eps_one
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in acosh(cosh(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.acosh(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.acosh(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"acosh({self._a!r})"


class ScalarAtanh(Expr[AnyExpr]):
    __slots__: tuple[str, ...] = ("_a",)
    _a: AnyExpr

    def __init__(self, a: AnyExpr) -> None:
        self._a = a

    @property
    @override
    def args(self) -> tuple[AnyExpr]:
        return (self._a,)

    @override
    def with_args(self, a: AnyExpr) -> "ScalarAtanh":
        return ScalarAtanh(a)

    @override
    def constant_fold(self, dtype: np.dtype[F]) -> F | AnyExpr:
        return ScalarFoldedConstant.constant_fold_unary(
            self._a,
            dtype,
            np.atanh,
            ScalarAtanh,
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return np.atanh(self._a.eval(Xs, late_bound))

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate arg and atanh(arg)
        arg = self._a
        argv = arg.eval(Xs, late_bound)
        exprv = np.atanh(argv)

        one_eps = np.nextafter(Xs.dtype.type(1), Xs.dtype.type(2))

        # apply the inverse function to get the bounds on arg
        # atanh(...) is NaN when abs(...) > 1 and can then take any value > 1
        # if arg_lower == argv and argv == -0.0, we need to guarantee that
        #  arg_lower is also -0.0, same for arg_upper
        arg_lower: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.tanh(expr_lower)
        )
        arg_lower[np.greater(argv, 1)] = one_eps
        arg_lower[np.less(argv, -1)] = -np.inf
        arg_lower = _ensure_array(_minimum_zero_sign_sensitive(argv, arg_lower))

        arg_upper: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            np.tanh(expr_upper)
        )
        arg_upper[np.greater(argv, 1)] = np.inf
        arg_upper[np.less(argv, -1)] = -one_eps
        arg_upper = _ensure_array(_maximum_zero_sign_sensitive(argv, arg_upper))

        # we need to force argv if expr_lower == expr_upper
        np.copyto(arg_lower, argv, where=(expr_lower == expr_upper), casting="no")
        np.copyto(arg_upper, argv, where=(expr_lower == expr_upper), casting="no")

        # handle rounding errors in atanh(tanh(...)) early
        arg_lower = guarantee_arg_within_expr_bounds(
            lambda arg_lower: np.atanh(arg_lower),
            exprv,
            argv,
            arg_lower,
            expr_lower,
            expr_upper,
        )
        arg_upper = guarantee_arg_within_expr_bounds(
            lambda arg_upper: np.atanh(arg_upper),
            exprv,
            argv,
            arg_upper,
            expr_lower,
            expr_upper,
        )

        return arg.compute_data_bounds(
            arg_lower,
            arg_upper,
            Xs,
            late_bound,
        )

    @override
    def __repr__(self) -> str:
        return f"atanh({self._a!r})"



"""
Same value safeguard.
"""

__all__ = ["SameValueSafeguard"]

from collections.abc import Set
from typing import ClassVar, Literal

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ...utils._compat import _ensure_array, _logical_and
from ...utils.bindings import Bindings, Parameter
from ...utils.cast import as_bits, from_total_order, lossless_cast, to_total_order
from ...utils.error import TypeCheckError, ctx
from ...utils.intervals import Interval, IntervalUnion, Lower, Maximum, Minimum, Upper
from ...utils.typing import JSON, S, T
from .abc import PointwiseSafeguard


class SameValueSafeguard(PointwiseSafeguard):
    """
    The `SameValueSafeguard` guarantees that if an element has a special
    `value` in the input, that element also has bitwise the same value in the
    decompressed output.

    This safeguard can be used for preserving e.g. zero values, missing values,
    pre-computed extreme values, or any other value of importance.

    By default, elements that do *not* have the special `value` in the input
    may still have the value in the output. Enabling the `exclusive` flag
    enforces that an element in the output only has the special `value` if and
    only if it also has the `value` in the input, e.g. to ensure that only
    missing values in the input have the missing value bitpattern in the
    output.

    Beware that +0.0 and -0.0 are semantically equivalent in floating-point but
    have different bitwise patterns. To preserve both, two same value
    safeguards are needed, one for each bitpattern.

    Parameters
    ----------
    value : int | float | str | Parameter
        The value of or the late-bound parameter name for the certain `value`
        that is preserved by this safeguard. Literal values are (unsafely) cast
        to the data dtype before binary comparison.
    exclusive : bool
        If [`True`][True], non-`value` elements in the data stay non-`value`
        after applying corrections. If [`False`][False], non-`value` values may
        have the `value` after applying corrections.

    Raises
    ------
    TypeCheckError
        if any parameter has the wrong type.
    """

    __slots__: tuple[str, ...] = ("_value", "_exclusive")
    _value: int | float | Parameter
    _exclusive: bool

    kind: ClassVar[str] = "same"

    def __init__(
        self, value: int | float | str | Parameter, *, exclusive: bool = False
    ) -> None:
        with ctx.safeguard(self):
            with ctx.parameter("value"):
                TypeCheckError.check_instance_or_raise(
                    value, int | float | str | Parameter
                )
                if isinstance(value, Parameter):
                    self._value = value
                elif isinstance(value, str):
                    self._value = Parameter(value)
                else:
                    self._value = value

            with ctx.parameter("exclusive"):
                TypeCheckError.check_instance_or_raise(exclusive, bool)
                self._exclusive = exclusive

    @property
    @override
    def late_bound(self) -> Set[Parameter]:
        """
        The set of late-bound parameters that this safeguard has.

        Late-bound parameters are only bound when checking and applying the
        safeguard, in contrast to the normal early-bound parameters that are
        configured during safeguard initialisation.

        Late-bound parameters can be used for parameters that depend on the
        specific data that is to be safeguarded.
        """

        return (
            frozenset([self._value])
            if isinstance(self._value, Parameter)
            else frozenset()
        )

    @override
    def check_pointwise(
        self,
        data: np.ndarray[S, np.dtype[T]],
        prediction: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Check which elements preserve the special `value` from the `data` to
        the `prediction` array.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Original data array, relative to which the `prediction` is checked.
        prediction : np.ndarray[S, np.dtype[T]]
            Prediction for the `data` array.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only check at data points where the condition is [`True`][True].

        Returns
        -------
        ok : np.ndarray[S, np.dtype[np.bool]]
            Pointwise, `True` if the check succeeded for this element.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `value` is late-bound but its late-bound parameter is not in
            `late_bound`.
        ValueError
            if the late-bound `value` could not be broadcast to the `data`'s
            shape.
        TypeError
            if the `value` is floating-point but the `data` is integer.
        ValueError
            if not all `value`s could not be losslessly converted to the
            `data`'s type.
        """

        with ctx.safeguard(self), ctx.parameter("value"):
            value: np.ndarray[tuple[()] | S, np.dtype[T]] = (
                late_bound.resolve_ndarray_with_lossless_cast(
                    self._value,
                    data.shape,
                    data.dtype,
                )
                if isinstance(self._value, Parameter)
                else lossless_cast(self._value, data.dtype)
            )
        value_bits: np.ndarray[tuple[()] | S, np.dtype[np.unsignedinteger]] = as_bits(
            value
        )

        data_bits: np.ndarray[S, np.dtype[np.unsignedinteger]] = as_bits(data)
        prediction_bits: np.ndarray[S, np.dtype[np.unsignedinteger]] = as_bits(
            prediction
        )

        ok: np.ndarray[S, np.dtype[np.bool]]
        if self._exclusive:
            # value if and only if where value
            ok = (data_bits == value_bits) == (prediction_bits == value_bits)
        else:
            # value must stay value, everything else can be arbitrary
            ok = (data_bits != value_bits) | (prediction_bits == value_bits)
        ok = _ensure_array(ok)
        if where is not True:
            ok[~where] = True

        return ok

    @override
    def compute_safe_intervals(
        self,
        data: np.ndarray[S, np.dtype[T]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> IntervalUnion[T, int, int]:
        """
        Compute the intervals in which the same value guarantee is upheld with
        respect to the `data`.

        Parameters
        ----------
        data : np.ndarray[S, np.dtype[T]]
            Data for which the safe intervals should be computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the safe intervals at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        intervals : IntervalUnion[T, int, int]
            Union of intervals in which the same value guarantee is upheld.

        Raises
        ------
        LateBoundParameterResolutionError
            if the `value` is late-bound but its late-bound parameter is not in
            `late_bound`.
        ValueError
            if the late-bound `value` could not be broadcast to the `data`'s
            shape.
        TypeError
            if the `value` is floating-point but the `data` is integer.
        ValueError
            if not all `value`s could not be losslessly converted to the
            `data`'s type.
        """

        with ctx.safeguard(self), ctx.parameter("value"):
            valuef: np.ndarray[tuple[()] | tuple[int], np.dtype[T]] = (
                late_bound.resolve_ndarray_with_lossless_cast(
                    self._value,
                    data.shape,
                    data.dtype,
                ).flatten()
                if isinstance(self._value, Parameter)
                else lossless_cast(self._value, data.dtype)
            )
        valuef_bits: np.ndarray[
            tuple[()] | tuple[int], np.dtype[np.unsignedinteger]
        ] = as_bits(valuef)

        dataf: np.ndarray[tuple[int], np.dtype[T]] = data.flatten()
        dataf_bits: np.ndarray[tuple[int], np.dtype[np.unsignedinteger]] = as_bits(
            dataf
        )

        wheref: Literal[True] | np.ndarray[tuple[int], np.dtype[np.bool]] = (
            True if where is True else where.flatten()
        )

        valid = Interval.empty_like(dataf)

        if not self._exclusive:
            # preserve value elements exactly, do not constrain other elements
            valid = Interval.full_like(dataf)
            Lower(valuef) <= valid[dataf_bits == valuef_bits] <= Upper(valuef)
            return valid.preserve_only_where(wheref).into_union()

        valuef_total: np.ndarray[
            tuple[()] | tuple[int], np.dtype[np.unsignedinteger]
        ] = to_total_order(valuef)

        total_min = np.array(np.iinfo(valuef_total.dtype).min, dtype=valuef_total.dtype)
        total_max = np.array(np.iinfo(valuef_total.dtype).max, dtype=valuef_total.dtype)

        valid_below = Interval.empty_like(dataf)
        valid_above = Interval.empty_like(dataf)

        Lower(valuef) <= valid_below[dataf_bits == valuef_bits] <= Upper(valuef)

        with np.errstate(over="ignore", under="ignore"):
            below_upper = _ensure_array(from_total_order(valuef_total - 1, data.dtype))
            above_lower = _ensure_array(from_total_order(valuef_total + 1, data.dtype))

        # non-value elements must exclude value from their interval,
        #  leading to a union of two intervals, below and above value
        Minimum <= valid_below[
            (dataf_bits != valuef_bits) & (valuef_total > total_min)
        ] <= Upper(below_upper)

        Lower(above_lower) <= valid_above[
            (dataf_bits != valuef_bits) & (valuef_total < total_max)
        ] <= Maximum

        return valid_below.union(valid_above).preserve_only_where(wheref)

    @override
    def compute_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the footprint of the `foot` array, e.g. for expanding data
        points into the pointwise checks that they contribute to.

        The footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the footprint at pointwise checks where the condition
            is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def compute_inverse_footprint(
        self,
        foot: np.ndarray[S, np.dtype[np.bool]],
        *,
        late_bound: Bindings,
        where: Literal[True] | np.ndarray[S, np.dtype[np.bool]] = True,
    ) -> np.ndarray[S, np.dtype[np.bool]]:
        """
        Compute the inverse footprint of the `foot` array, e.g. for expanding
        pointwise check fails into the points that could have contributed to
        the failures.

        The inverse footprint is equivalent to `foot & where`.

        Parameters
        ----------
        foot : np.ndarray[S, np.dtype[np.bool]]
            Array for which the inverse footprint is computed.
        late_bound : Bindings
            Bindings for late-bound parameters, including for this safeguard.
        where : Literal[True] | np.ndarray[S, np.dtype[np.bool]]
            Only compute the inverse footprint at pointwise checks where the
            condition is [`True`][True].

        Returns
        -------
        print : np.ndarray[S, np.dtype[np.bool]]
            The inverse footprint of the `foot` array.
        """

        return _logical_and(foot, where)

    @override
    def get_config(self) -> dict[str, JSON]:
        """
        Returns the configuration of the safeguard.

        Returns
        -------
        config : dict[str, JSON]
            Configuration of the safeguard.
        """

        return dict(kind=type(self).kind, value=self._value, exclusive=self._exclusive)



from collections.abc import Mapping

import numpy as np
from typing_extensions import override  # MSPV 3.12

from ....utils._compat import (
    _broadcast_to,
    _ensure_array,
    _maximum_zero_sign_sensitive,
    _minimum_zero_sign_sensitive,
    _where,
)
from ....utils.bindings import Parameter
from ..bound import checked_data_bounds
from ..typing import F, Fi, Ns, Ps, np_sndarray
from .abc import AnyExpr, Expr
from .constfold import ScalarFoldedConstant


class ScalarWhere(Expr[AnyExpr, AnyExpr, AnyExpr]):
    __slots__: tuple[str, ...] = ("_condition", "_a", "_b")
    _condition: AnyExpr
    _a: AnyExpr
    _b: AnyExpr

    def __init__(self, condition: AnyExpr, a: AnyExpr, b: AnyExpr):
        self._condition = condition
        self._a = a
        self._b = b

    @property
    @override
    def args(self) -> tuple[AnyExpr, AnyExpr, AnyExpr]:
        return (self._condition, self._a, self._b)

    @override
    def with_args(self, condition: AnyExpr, a: AnyExpr, b: AnyExpr) -> "ScalarWhere":
        return ScalarWhere(condition, a, b)

    @override  # type: ignore
    def eval_has_data(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[np.bool]]:
        has_data = self._condition.eval_has_data(Xs, late_bound)
        has_data |= _where(
            self._condition.eval(Xs, late_bound) != 0,
            self._a.eval_has_data(Xs, late_bound),
            self._b.eval_has_data(Xs, late_bound),
        )
        return has_data

    @override
    def constant_fold(self, dtype: np.dtype[Fi]) -> Fi | AnyExpr:
        cond = self._condition.constant_fold(dtype)
        a = self._a.constant_fold(dtype)
        b = self._b.constant_fold(dtype)

        if not isinstance(cond, Expr):
            if cond != 0 and not isinstance(a, Expr):
                return a

            if cond == 0 and not isinstance(b, Expr):
                return b

        return ScalarWhere(
            ScalarFoldedConstant.from_folded(cond),
            ScalarFoldedConstant.from_folded(a),
            ScalarFoldedConstant.from_folded(b),
        )

    @override
    def eval(
        self,
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> np.ndarray[tuple[Ps], np.dtype[F]]:
        return _where(
            self._condition.eval(Xs, late_bound) != 0,
            self._a.eval(Xs, late_bound),
            self._b.eval(Xs, late_bound),
        )

    @checked_data_bounds
    @override
    def compute_data_bounds_unchecked(
        self,
        expr_lower: np.ndarray[tuple[Ps], np.dtype[F]],
        expr_upper: np.ndarray[tuple[Ps], np.dtype[F]],
        Xs: np_sndarray[Ps, Ns, np.dtype[F]],
        late_bound: Mapping[Parameter, np_sndarray[Ps, Ns, np.dtype[F]]],
    ) -> tuple[
        np_sndarray[Ps, Ns, np.dtype[F]],
        np_sndarray[Ps, Ns, np.dtype[F]],
    ]:
        # evaluate the condition, a, and b
        cond, a, b = self._condition, self._a, self._b
        condv: np.ndarray[tuple[Ps], np.dtype[F]] = _ensure_array(
            cond.eval(Xs, late_bound)
        )
        condvb_Ps: np.ndarray[tuple[Ps], np.dtype[np.bool]] = condv != 0
        condvb_Ns: np_sndarray[Ps, Ns, np.dtype[np.bool]] = _broadcast_to(
            _ensure_array(condvb_Ps).reshape(Xs.shape[:1] + (1,) * (Xs.ndim - 1)),
            Xs.shape,
        )
        av = a.eval(Xs, late_bound)
        bv = b.eval(Xs, late_bound)

        Xs_lower: np_sndarray[Ps, Ns, np.dtype[F]] = np.full(
            Xs.shape, Xs.dtype.type(-np.inf)
        )
        Xs_upper: np_sndarray[Ps, Ns, np.dtype[F]] = np.full(
            Xs.shape, Xs.dtype.type(np.inf)
        )

        if cond.has_data:
            # for simplicity, we assume that the condition must always evaluate
            #  to the same boolean when compared to 0
            cond_lower: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                condv.shape, Xs.dtype.type(-np.inf)
            )
            cond_upper: np.ndarray[tuple[Ps], np.dtype[F]] = np.full(
                condv.shape, Xs.dtype.type(np.inf)
            )

            # zero condition values must remain zero
            cond_lower[condv == 0] = -0.0
            cond_upper[condv == 0] = +0.0

            smallest_subnormal = np.finfo(Xs.dtype).smallest_subnormal

            # non-zero condition values must remain non-zero
            # TODO: an interval union could represent that the two disjoint
            #       intervals in the future
            cond_lower[condv > 0] = smallest_subnormal
            cond_upper[condv < 0] = -smallest_subnormal

            cl, cu = cond.compute_data_bounds(cond_lower, cond_upper, Xs, late_bound)
            Xs_lower = _maximum_zero_sign_sensitive(Xs_lower, cl)
            Xs_upper = _minimum_zero_sign_sensitive(Xs_upper, cu)

        if np.any(condvb_Ps) and a.has_data:
            # pass on the data bounds to a but only use its bounds on Xs if
            #  chosen by the condition
            al, au = a.compute_data_bounds(
                _where(condvb_Ps, expr_lower, av),
                _where(condvb_Ps, expr_upper, av),
                Xs,
                late_bound,
            )

            # combine the data bounds
            np.copyto(
                Xs_lower,
                _maximum_zero_sign_sensitive(Xs_lower, al),
                where=condvb_Ns,
                casting="no",
            )
            np.copyto(
                Xs_upper,
                _minimum_zero_sign_sensitive(Xs_upper, au),
                where=condvb_Ns,
                casting="no",
            )

        if (not np.all(condvb_Ps)) and b.has_data:
            # pass on the data bounds to b but only use its bounds on Xs if
            #  chosen by the condition
            bl, bu = b.compute_data_bounds(
                _where(condvb_Ps, bv, expr_lower),
                _where(condvb_Ps, bv, expr_upper),
                Xs,
                late_bound,
            )

            # combine the data bounds
            np.copyto(
                Xs_lower,
                _maximum_zero_sign_sensitive(Xs_lower, bl),
                where=~condvb_Ns,
                casting="no",
            )
            np.copyto(
                Xs_upper,
                _minimum_zero_sign_sensitive(Xs_upper, bu),
                where=~condvb_Ns,
                casting="no",
            )

        return Xs_lower, Xs_upper

    @override
    def __repr__(self) -> str:
        return f"where({self._condition!r}, {self._a!r}, {self._b!r})"



import itertools
from contextlib import contextmanager

from sly import Parser

from ...utils.bindings import Parameter
from ...utils.error import ctx
from .expr.abc import AnyExpr
from .expr.abs import ScalarAbs
from .expr.addsub import ScalarAdd, ScalarLeftAssociativeSum, ScalarSubtract
from .expr.array import Array
from .expr.classification import ScalarIsFinite, ScalarIsInf, ScalarIsNaN
from .expr.combinators import ScalarAll, ScalarAny, ScalarNot
from .expr.comparison import (
    ScalarEqual,
    ScalarGreater,
    ScalarGreaterEqual,
    ScalarLess,
    ScalarLessEqual,
    ScalarNotEqual,
)
from .expr.data import Data, LateBoundConstant
from .expr.divmul import ScalarDivide, ScalarMultiply
from .expr.finite_difference import (
    FiniteDifference,
    ScalarSymmetricModulo,
    finite_difference_coefficients,
    finite_difference_offsets,
)
from .expr.group import Group
from .expr.hyperbolic import (
    ScalarAcosh,
    ScalarAsinh,
    ScalarAtanh,
    ScalarCosh,
    ScalarSinh,
    ScalarTanh,
)
from .expr.literal import Euler, Number, Pi
from .expr.logexp import Exponential, Logarithm, ScalarExp, ScalarLog, ScalarLogWithBase
from .expr.neg import ScalarNegate
from .expr.power import ScalarPower
from .expr.reciprocal import ScalarReciprocal
from .expr.round import ScalarCeil, ScalarFloor, ScalarRoundTiesEven, ScalarTrunc
from .expr.sign import ScalarSign
from .expr.square import ScalarSqrt, ScalarSquare
from .expr.trigonometric import (
    ScalarAcos,
    ScalarAsin,
    ScalarAtan,
    ScalarCos,
    ScalarSin,
    ScalarTan,
)
from .expr.where import ScalarWhere
from .lexer import QoILexer


class QoIParser(Parser):
    __slots__: tuple[str, ...] = ("_x", "_X", "_I", "_vars", "_text")
    _x: Data
    _X: None | Array
    _I: None | tuple[int, ...]
    _vars: dict[Parameter, AnyExpr]
    _text: str

    def __init__(
        self,
        *,
        x: Data,
        X: None | Array,
        I: None | tuple[int, ...],  # noqa: E741
    ):
        self._x = x
        self._X = X
        self._I = I
        self._vars = dict()

    def parse2(self, text: str, tokens):
        self._text = text
        tokens, tokens2 = itertools.tee(tokens)

        if next(tokens2, None) is None:
            raise (
                SyntaxError("expression must not be empty", ("<qoi>", None, None, None))
                | ctx
            )

        return super().parse(tokens)

    tokens = QoILexer.tokens

    # === operator precedence and associativity ===
    precedence = (
        # lowest precedence: comparisons
        ("nonassoc", LESS, LESS_EQUAL, EQUAL, NOT_EQUAL, GREATER_EQUAL, GREATER),  # type: ignore[name-defined]  # noqa: F821
        ("left", PLUS, MINUS),  # type: ignore[name-defined]  # noqa: F821
        ("left", TIMES, DIVIDE),  # type: ignore[name-defined]  # noqa: F821
        ("right", UPLUS, UMINUS),  # type: ignore[name-defined]  # noqa: F821
        ("right", POWER),  # type: ignore[name-defined]  # noqa: F821
        ("left", INDEX, TRANSPOSE),  # type: ignore[name-defined]  # noqa: F821
        # highest precedence: array indexing and transpose
        # help array indexing by giving the opening `[` an even higher precedence
        ("left", LBRACK),  # type: ignore[name-defined]  # noqa: F821
    )

    # === grammar rules ===

    # top-level: qoi := expr | { assign } return expr;
    @_("expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def qoi(self, p):  # noqa: F811
        self.assert_or_error(
            not isinstance(p.expr, Array),
            p,
            lambda: f"expression must be a scalar but is an array expression of shape {p.expr.shape}",
        )
        self.assert_or_error(p.expr.has_data, p, "expression must not be constant")
        return p.expr

    @_("many_assign return_expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def qoi(self, p):  # noqa: F811
        return p.return_expr

    @_("RETURN expr SEMI")  # type: ignore[name-defined]  # noqa: F821
    def return_expr(self, p):
        self.assert_or_error(
            not isinstance(p.expr, Array),
            p,
            lambda: f"return expression must be a scalar but is an array expression of shape {p.expr.shape}",
        )
        self.assert_or_error(
            p.expr.has_data, p, "return expression must not be constant"
        )
        return p.expr

    # variable assignment: assign := V["id"] = expr;
    @_("VS LBRACK quotedparameter RBRACK ASSIGN expr SEMI")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def assign(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is None, p, "stencil QoI variables use upper-case `V`"
        )
        self.assert_or_error(
            not p.quotedparameter.is_builtin,
            p,
            "variable name must be built-in (start with `$`)",
        )
        self.assert_or_error(
            p.quotedparameter not in self._vars,
            p,
            f'cannot override already-defined variable v["{p.quotedparameter}"]',
        )
        self._vars[p.quotedparameter] = Array.map(Group, p.expr)

    @_("VA LBRACK quotedparameter RBRACK ASSIGN expr SEMI")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def assign(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is not None, p, "pointwise QoI variables use lower-case `v`"
        )
        self.assert_or_error(
            not p.quotedparameter.is_builtin,
            p,
            "variable name must be built-in (start with `$`)",
        )
        self.assert_or_error(
            p.quotedparameter not in self._vars,
            p,
            f'cannot override already-defined variable V["{p.quotedparameter}"]',
        )
        self._vars[p.quotedparameter] = Array.map(Group, p.expr)

    @_("ID ASSIGN expr SEMI")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def assign(self, p):  # noqa: F811
        self.raise_error(
            p,
            f'cannot assign to identifier `{p.ID}`, assign to a variable {"v" if self._X is None else "V"}["{p.ID}"] instead',
        )

    @_("assign many_assign")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def many_assign(self, p):  # noqa: F811
        pass

    @_("empty")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def many_assign(self, p):  # noqa: F811
        pass

    # integer literal (non-expression)
    @_("INTEGER")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def integer(self, p):  # noqa: F811
        return p.INTEGER

    @_("PLUS INTEGER")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def integer(self, p):  # noqa: F811
        return p.INTEGER

    @_("MINUS INTEGER")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def integer(self, p):  # noqa: F811
        return -p.INTEGER

    # expressions

    # integer and floating-point literals
    @_("INTEGER")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Number.from_symbolic_int(p.INTEGER)

    @_("FLOAT")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Number(p.FLOAT)

    @_("INF")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Number("inf")

    @_("NAN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Number("nan")

    # array literal
    @_("LBRACK expr many_comma_expr RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(
            p, lambda err: f"invalid array literal: {err}", exception=ValueError
        ):
            return Array(*([p.expr] + p.many_comma_expr))

    @_("LBRACK RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.raise_error(p, "invalid empty array literal")

    @_("comma_expr many_comma_expr")  # type: ignore[name-defined]  # noqa: F821
    def many_comma_expr(self, p):
        return [p.comma_expr] + p.many_comma_expr

    @_("COMMA")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def many_comma_expr(self, p):  # noqa: F811
        return []

    @_("empty")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def many_comma_expr(self, p):  # noqa: F811
        return []

    @_("COMMA expr")  # type: ignore[name-defined]  # noqa: F821
    def comma_expr(self, p):
        return p.expr

    # unary operators (positive, negative):
    #  expr := OP expr
    @_("PLUS expr %prec UPLUS")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return p.expr

    @_("MINUS expr %prec UMINUS")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarNegate, p.expr)

    # binary operators (add, subtract, multiply, divide, power):
    #  expr := expr OP expr
    @_("expr PLUS expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarAdd, p.expr0, p.expr1)

    @_("expr MINUS expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarSubtract, p.expr0, p.expr1)

    @_("expr TIMES expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarMultiply, p.expr0, p.expr1)

    @_("expr DIVIDE expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarDivide, p.expr0, p.expr1)

    @_("expr POWER expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarPower, p.expr0, p.expr1)

    # binary comparison operators
    #  expr := expr OP expr
    @_("expr LESS_EQUAL expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarLessEqual, p.expr0, p.expr1)

    @_("expr LESS expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarLess, p.expr0, p.expr1)

    @_("expr EQUAL expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarEqual, p.expr0, p.expr1)

    @_("expr NOT_EQUAL expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarNotEqual, p.expr0, p.expr1)

    @_("expr GREATER_EQUAL expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarGreaterEqual, p.expr0, p.expr1)

    @_("expr GREATER expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarGreater, p.expr0, p.expr1)

    # array transpose: expr := expr.T
    @_("expr TRANSPOSE")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            isinstance(p.expr, Array), p, "cannot transpose scalar non-array expression"
        )
        return p.expr.transpose()

    # group in parentheses
    @_("LPAREN expr RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(Group, p.expr)

    # optional trailing comma separator
    @_("COMMA")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def maybe_comma(self, p):  # noqa: F811
        pass

    @_("empty")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def maybe_comma(self, p):  # noqa: F811
        pass

    # constants: expr := e | pi
    @_("EULER")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Euler()

    @_("PI")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Pi()

    # data, late-bound constants, variables
    @_("XS")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return self._x

    @_("XA")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is not None,
            p,
            "data neighbourhood `X` is not available in pointwise QoIs, use pointwise `x` instead",
        )
        return self._X

    @_("CS LBRACK quotedparameter RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return LateBoundConstant.like(p.quotedparameter, self._x)

    @_("CA LBRACK quotedparameter RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is not None,
            p,
            "late-bound constant neighbourhood `C` is not available in pointwise QoIs, use pointwise `c` instead",
        )
        return Array.map(
            lambda e: LateBoundConstant.like(p.quotedparameter, e),
            self._X,
        )

    @_("VS LBRACK quotedparameter RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is None, p, "stencil QoI variables use upper-case `V`"
        )
        self.assert_or_error(
            not p.quotedparameter.is_builtin,
            p,
            "variable name must not be built-in (start with `$`)",
        )
        self.assert_or_error(
            p.quotedparameter in self._vars,
            p,
            f'undefined variable v["{p.quotedparameter}"]',
        )
        return self._vars[p.quotedparameter]

    @_("VA LBRACK quotedparameter RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is not None, p, "pointwise QoI variables use lower-case `v`"
        )
        self.assert_or_error(
            not p.quotedparameter.is_builtin,
            p,
            "variable name must not be built-in (start with `$`)",
        )
        self.assert_or_error(
            p.quotedparameter in self._vars,
            p,
            f'undefined variable V["{p.quotedparameter}"]',
        )
        return self._vars[p.quotedparameter]

    @_("STRING")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def quotedparameter(self, p):  # noqa: F811
        with self.with_error_context(
            p, f'invalid quoted parameter "{p.STRING}": must be a valid identifier'
        ):
            return Parameter(p.STRING)

    # array indexing
    @_("expr LBRACK IDX RBRACK %prec INDEX")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._I is not None, p, "index `I` is not available in pointwise QoIs"
        )
        self.assert_or_error(
            isinstance(p.expr, Array), p, "cannot index scalar non-array expression"
        )
        with self.with_error_context(p, lambda err: f"{err}", exception=IndexError):
            return p.expr.index(self._I)

    @_("expr LBRACK index_ many_comma_index RBRACK %prec INDEX")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            isinstance(p.expr, Array), p, "cannot index scalar non-array expression"
        )
        with self.with_error_context(p, lambda err: f"{err}", exception=IndexError):
            return p.expr.index(tuple([p.index_] + p.many_comma_index))

    @_("comma_index many_comma_index")  # type: ignore[name-defined]  # noqa: F821
    def many_comma_index(self, p):
        return [p.comma_index] + p.many_comma_index

    @_("COMMA")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def many_comma_index(self, p):  # noqa: F811
        return []

    @_("empty")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def many_comma_index(self, p):  # noqa: F811
        return []

    @_("COMMA index_")  # type: ignore[name-defined]  # noqa: F821
    def comma_index(self, p):
        return p.index_

    @_("integer_expr")  # type: ignore[name-defined]  # noqa: F821
    def index_(self, p):
        return p.integer_expr

    @_("maybe_integer_expr COLON maybe_integer_expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def index_(self, p):  # noqa: F811
        return slice(p.maybe_integer_expr0, p.maybe_integer_expr1, None)

    @_("maybe_integer_expr COLON maybe_integer_expr COLON maybe_integer_expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def index_(self, p):  # noqa: F811
        return slice(
            p.maybe_integer_expr0, p.maybe_integer_expr1, p.maybe_integer_expr2
        )

    @_("integer_expr")  # type: ignore[name-defined]  # noqa: F821
    def maybe_integer_expr(self, p):
        return p.integer_expr

    @_("empty")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def maybe_integer_expr(self, p):  # noqa: F811
        return None

    @_("expr")  # type: ignore[name-defined]  # noqa: F821
    def integer_expr(self, p):
        self.assert_or_error(
            isinstance(p.expr, Number) and p.expr.as_int() is not None,
            p,
            "cannot index by non-integer expression",
        )
        return p.expr.as_int()

    @_("IDX LBRACK index_ many_comma_index RBRACK")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._I is not None, p, "index `I` is not available in pointwise QoIs"
        )
        idx = Array(*tuple(Number.from_symbolic_int(i) for i in self._I))
        with self.with_error_context(
            p,
            lambda err: f"{err} for `I`, the 1D per-axis stencil centre index array",
            exception=IndexError,
        ):
            return idx.index(tuple([p.index_] + p.many_comma_index))

    # functions

    # logarithms and exponentials
    @_("LN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(lambda e: ScalarLog(Logarithm.ln, e), p.expr)

    @_("LOG2 LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(lambda e: ScalarLog(Logarithm.log2, e), p.expr)

    @_("LOG10 LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(lambda e: ScalarLog(Logarithm.log10, e), p.expr)

    @_("LOG LPAREN expr COMMA BASE ASSIGN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarLogWithBase, p.expr0, p.expr1)

    @_("EXP LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(lambda e: ScalarExp(Exponential.exp, e), p.expr)

    @_("EXP2 LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(lambda e: ScalarExp(Exponential.exp2, e), p.expr)

    @_("EXP10 LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(lambda e: ScalarExp(Exponential.exp10, e), p.expr)

    # exponentiation
    @_("SQRT LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarSqrt, p.expr)

    @_("SQUARE LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarSquare, p.expr)

    @_("RECIPROCAL LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarReciprocal, p.expr)

    # absolute value
    @_("ABS LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAbs, p.expr)

    # sign and rounding
    @_("SIGN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarSign, p.expr)

    @_("FLOOR LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarFloor, p.expr)

    @_("CEIL LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarCeil, p.expr)

    @_("TRUNC LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarTrunc, p.expr)

    @_("ROUND_TIES_EVEN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarRoundTiesEven, p.expr)

    # trigonometric
    @_("SIN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarSin, p.expr)

    @_("COS LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarCos, p.expr)

    @_("TAN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarTan, p.expr)

    @_("ASIN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAsin, p.expr)

    @_("ACOS LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAcos, p.expr)

    @_("ATAN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAtan, p.expr)

    # hyperbolic
    @_("SINH LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarSinh, p.expr)

    @_("COSH LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarCosh, p.expr)

    @_("TANH LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarTanh, p.expr)

    @_("ASINH LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAsinh, p.expr)

    @_("ACOSH LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAcosh, p.expr)

    @_("ATANH LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarAtanh, p.expr)

    # classification
    @_("ISFINITE LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarIsFinite, p.expr)

    @_("ISINF LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarIsInf, p.expr)

    @_("ISNAN LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarIsNaN, p.expr)

    # combinators
    @_("NOT LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        return Array.map(ScalarNot, p.expr)

    @_(  # type: ignore[name-defined, no-redef]  # noqa: F821
        "ALL LPAREN expr maybe_comma RPAREN"
    )
    def expr(self, p):  # noqa: F811
        expr = p.expr
        self.assert_or_error(
            isinstance(expr, Array) and expr.size >= 2,
            p,
            "`all` expr must be an array expression with at least two elements",
        )
        a, b, *cs = expr.flatlist()

        # must not fail since flat Array elements cannot be arrays themselves
        assert not isinstance(a, Array)
        assert not isinstance(b, Array)
        assert not any(isinstance(c, Array) for c in cs)

        return ScalarAll(a, b, *cs)

    @_(  # type: ignore[name-defined, no-redef]  # noqa: F821
        "ANY LPAREN expr maybe_comma RPAREN"
    )
    def expr(self, p):  # noqa: F811
        expr = p.expr
        self.assert_or_error(
            isinstance(expr, Array) and expr.size >= 2,
            p,
            "`any` expr must be an array expression with at least two elements",
        )
        a, b, *cs = expr.flatlist()

        # must not fail since flat Array elements cannot be arrays themselves
        assert not isinstance(a, Array)
        assert not isinstance(b, Array)
        assert not any(isinstance(c, Array) for c in cs)

        return ScalarAny(a, b, *cs)

    @_("WHERE LPAREN expr COMMA expr COMMA expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.map(ScalarWhere, p.expr0, p.expr1, p.expr2)

    # array operations
    @_("SIZE LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            isinstance(p.expr, Array), p, "scalar non-array expression has no size"
        )
        return Number.from_symbolic_int(p.expr.size)

    @_("SHAPE LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            isinstance(p.expr, Array), p, "scalar non-array expression has no shape"
        )
        return Array(*tuple(Number.from_symbolic_int(s) for s in p.expr.shape))

    @_("SUM LPAREN expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            isinstance(p.expr, Array), p, "cannot sum over scalar non-array expression"
        )
        return p.expr.sum()

    @_("MATMUL LPAREN expr COMMA expr maybe_comma RPAREN")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            isinstance(p.expr0, Array),
            p,
            "cannot matmul(a, b) with scalar non-array parameter a",
        )
        self.assert_or_error(
            isinstance(p.expr1, Array),
            p,
            "cannot matmul(a, b) with scalar non-array parameter b",
        )
        with self.with_error_context(p, lambda err: f"{err}", exception=ValueError):
            return Array.matmul(p.expr0, p.expr1)

    # finite difference
    @_(  # type: ignore[name-defined, no-redef]  # noqa: F821
        "FINITE_DIFFERENCE LPAREN expr COMMA ORDER ASSIGN integer COMMA ACCURACY ASSIGN integer COMMA TYPE ASSIGN integer COMMA AXIS ASSIGN integer finite_difference_grid_spacing finite_difference_grid_period RPAREN"
    )
    def expr(self, p):  # noqa: F811
        self.assert_or_error(
            self._X is not None,
            p,
            "`finite_difference` is not available in pointwise QoIs",
        )

        expr = p.expr
        self.assert_or_error(
            p.expr.has_data,
            p,
            f"`finite_difference` expr must reference the data `{'x' if self._X is None else 'X'}`",
        )
        self.assert_or_error(
            not isinstance(p.expr, Array),
            p,
            "`finite_difference` expr must be a scalar array element expression, e.g. the centre value, not an array",
        )

        order = p.integer0
        self.assert_or_error(
            order >= 0, p, "`finite_difference` order must be non-negative"
        )

        accuracy = p.integer1
        self.assert_or_error(
            accuracy > 0, p, "`finite_difference` accuracy must be positive"
        )

        TYPES: dict[int, FiniteDifference] = {
            -1: FiniteDifference.backwards,
            0: FiniteDifference.central,
            1: FiniteDifference.forward,
        }
        type = p.integer2
        self.assert_or_error(
            type in TYPES,
            p,
            "`finite_difference` type must be 1 (forward), 0 (central), or -1 (backward)",
        )
        type = TYPES[type]
        if type == FiniteDifference.central:
            self.assert_or_error(
                accuracy % 2 == 0,
                p,
                "`finite_difference` accuracy must be even for a central finite difference",
            )

        axis = p.integer3
        self.assert_or_error(
            axis >= -len(self._X.shape) and axis < len(self._X.shape),
            p,
            f"`finite_difference` axis must be in range of the dimension {len(self._X.shape)} of the neighbourhood",
        )

        offsets = finite_difference_offsets(type, order, accuracy)

        grid_period = p.finite_difference_grid_period
        delta_transform = (
            (lambda e: e)
            if grid_period is None
            else (
                lambda e: Array.map(lambda f: ScalarSymmetricModulo(f, grid_period), e)
            )
        )

        if "spacing" in p.finite_difference_grid_spacing:
            grid_spacing = Group(p.finite_difference_grid_spacing["spacing"])

            coefficients = finite_difference_coefficients(
                order,
                tuple(
                    ScalarMultiply(Number.from_symbolic_int(o), grid_spacing)
                    for o in offsets
                ),
                lambda a: a,
                delta_transform=delta_transform,
            )
        else:
            grid_centre = Group(p.finite_difference_grid_spacing["centre"])

            coefficients = finite_difference_coefficients(
                order,
                tuple(grid_centre.apply_array_element_offset(axis, o) for o in offsets),
                lambda a: Group(ScalarSubtract(a, grid_centre)),
                delta_transform=delta_transform,
            )

        terms = [
            ScalarMultiply(expr.apply_array_element_offset(axis, o), c)
            for o, c in zip(offsets, coefficients)
        ]
        # even order=0 produces at least one term
        assert len(terms) > 0
        sum_ = ScalarLeftAssociativeSum(*terms)

        required_axis_before = 0
        required_axis_after = 0

        for idx in sum_.data_indices:
            required_axis_before = max(required_axis_before, self._I[axis] - idx[axis])
            required_axis_after = max(required_axis_after, idx[axis] - self._I[axis])

        self.assert_or_error(
            (required_axis_before <= self._I[axis])
            and (required_axis_after <= (self._X.shape[axis] - self._I[axis] - 1)),
            p,
            f"cannot compute the `finite_difference` on axis {axis} since the "
            "neighbourhood is insufficiently large: before should be at least "
            f"{required_axis_before} and after should be at least "
            f"{required_axis_after}",
        )

        assert not isinstance(sum_, Array)
        return Group(sum_)

    @_("COMMA GRID_SPACING ASSIGN expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def finite_difference_grid_spacing(self, p):  # noqa: F811
        self.assert_or_error(
            not p.expr.has_data,
            p,
            f"`finite_difference` grid_spacing must not reference the data `{'x' if self._X is None else 'X'}`",
        )
        self.assert_or_error(
            not isinstance(p.expr, Array),
            p,
            "`finite_difference` grid_spacing must be a constant scalar expression, not an array",
        )
        return dict(spacing=p.expr)

    @_("COMMA GRID_CENTRE ASSIGN expr")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def finite_difference_grid_spacing(self, p):  # noqa: F811
        self.assert_or_error(
            not p.expr.has_data,
            p,
            f"`finite_difference` grid_centre must not reference the data `{'x' if self._X is None else 'X'}`",
        )
        self.assert_or_error(
            not isinstance(p.expr, Array),
            p,
            "`finite_difference` grid_centre must be a constant scalar array element expression, not an array",
        )
        self.assert_or_error(
            len(p.expr.late_bound_constants) > 0,
            p,
            "`finite_difference` grid_centre must reference a late-bound constant",
        )
        return dict(centre=p.expr)

    @_("COMMA GRID_PERIOD ASSIGN expr maybe_comma")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def finite_difference_grid_period(self, p):  # noqa: F811
        self.assert_or_error(
            not p.expr.has_data,
            p,
            f"`finite_difference` grid_period must not reference the data `{'x' if self._X is None else 'X'}`",
        )
        self.assert_or_error(
            len(p.expr.late_bound_constants) == 0,
            p,
            "finite_difference grid_period must not reference late-bound constants",
        )
        self.assert_or_error(
            not isinstance(p.expr, Array),
            p,
            "`finite_difference` grid_period must be a constant scalar number, not an array",
        )
        return p.expr

    @_("maybe_comma")  # type: ignore[name-defined, no-redef]  # noqa: F821
    def finite_difference_grid_period(self, p):  # noqa: F811
        pass

    # empty rule
    @_("")  # type: ignore[name-defined]  # noqa: F821
    def empty(self, p):
        pass

    # === parser error handling ===
    def error(self, t):
        actions = self._lrtable.lr_action[self.state]
        options = ", ".join(QoILexer.token_to_name(a) for a in actions)
        oneof = " one of" if len(actions) > 1 else ""

        if t is None:
            raise (
                SyntaxError(
                    f"expected more input but found EOF\nexpected{oneof} {options}",
                    (
                        "<qoi>",
                        1 + self._text.count("\n"),
                        len(self._text) - self._text.rfind("\n"),
                        self._text[self._text.rfind("\n") + 1 :],
                    ),
                )
                | ctx
            )

        t_value = f'"{t.value}"' if t.type == "STRING" else t.value

        raise (
            SyntaxError(
                f"unexpected token `{t_value}`\nexpected{oneof} {options}",
                (
                    "<qoi>",
                    t.lineno,
                    self.find_column(t),
                    self._text.splitlines()[t.lineno - 1],
                ),
            )
            | ctx
        )

    def raise_error(self, t, message):
        raise (
            SyntaxError(
                message,
                (
                    "<qoi>",
                    t.lineno,
                    self.find_column(t),
                    self._text.splitlines()[t.lineno - 1],
                ),
            )
            | ctx
        )

    def assert_or_error(self, check, t, message):
        if not check:
            self.raise_error(t, message() if callable(message) else message)

    @contextmanager
    def with_error_context(self, t, message, exception=Exception):
        try:
            yield
        except exception as err:
            if callable(message):
                self.raise_error(t, message(err))
            else:
                self.raise_error(t, message)

    def find_column(self, token) -> int:
        last_cr = self._text.rfind("\n", 0, token.index)
        if last_cr < 0:
            last_cr = 0
        column = (token.index - last_cr) + 1
        return column
